import setuptools

setuptools.setup(
    name="ecoforesttest",
    version="0.0.2",
    author="",
    author_email="",
    description="",
    long_description="an interpretable data-driven macroeconomic forecasting method based on N-BEATS neural network",
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
